/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { type FieldKey, type MapTree, type SchemaAndPolicy } from "../core/index.js";
import { type CursorWithNode, typeNameSymbol, type NodeKeyManager } from "../feature-libraries/index.js";
import type { InsertableContent } from "./proxies.js";
import { FieldSchema, type ImplicitAllowedTypes, type TreeNodeSchema } from "./schemaTypes.js";
/**
 * Module notes:
 *
 * The flow of the below code is in terms of the structure of the input data. We then verify that the associated
 * schema is appropriate for that kind of data. This is fine while we have a 1:1 mapping of kind of input data to
 * the kind of schema we expect for it (e.g. an input that is an array always need to be associated with a sequence in
 * the schema). If/when we begin accepting kinds of input data that are ambiguous (e.g. accepting an input that is an
 * array of key/value tuples to instantiate a map) we may need to rethink the structure here to be based more on the
 * schema than on the input data.
 */
/**
 * Transforms an input {@link TypedNode} tree to a {@link MapTree}, and wraps the tree in a {@link CursorWithNode}.
 * @param data - The input tree to be converted.
 * @param allowedTypes - The set of types allowed by the parent context. Used to validate the input tree.
 * @param context - An optional context which, if present, will allow defaults to be created by {@link ContextualFieldProvider}s.
 * If absent, only defaults from {@link ConstantFieldProvider}s will be created.
 * @param schemaValidationPolicy - The stored schema and policy to be used for validation, if the policy says schema
 * validation should happen. If it does, the input tree will be validated against this schema + policy, and an error will
 * be thrown if the tree does not conform to the schema. If undefined, no validation against the stored schema is done.
 *
 * @returns A cursor (in nodes mode) for the mapped tree if the input data was defined. Otherwise, returns `undefined`.
 * @remarks The resulting tree will be populated with any defaults from {@link FieldProvider}s in the schema.
 */
export declare function cursorFromNodeData(data: InsertableContent, allowedTypes: ImplicitAllowedTypes, context?: NodeKeyManager, schemaValidationPolicy?: SchemaAndPolicy): CursorWithNode<MapTree>;
export declare function cursorFromNodeData(data: InsertableContent | undefined, allowedTypes: ImplicitAllowedTypes, context?: NodeKeyManager, schemaValidationPolicy?: SchemaAndPolicy): CursorWithNode<MapTree> | undefined;
/**
 * Transforms an input {@link InsertableContent} tree to an array of {@link MapTree}s, and wraps the tree in a {@link CursorWithNode}.
 * @param data - The input tree to be converted.
 * @param schema - Schema of the field with which the input `data` is associated.
 * @param context - An optional context which, if present, will allow defaults to be created by {@link ContextualFieldProvider}s.
 * If absent, only defaults from {@link ConstantFieldProvider}s will be created.
 * @param schemaValidationPolicy - The stored schema and policy to be used for validation, if the policy says schema
 * validation should happen. If it does, the input tree will be validated against this schema + policy, and an error will
 * be thrown if the tree does not conform to the schema. If undefined, no validation against the stored schema is done.
 * @remarks The resulting tree will be populated with any defaults from {@link FieldProvider}s in the schema.
 */
export declare function cursorFromFieldData(data: InsertableContent, schema: FieldSchema, context?: NodeKeyManager, schemaValidationPolicy?: SchemaAndPolicy): CursorWithNode<MapTree>;
export declare function cursorFromFieldData(data: InsertableContent | undefined, schema: FieldSchema, context?: NodeKeyManager, schemaValidationPolicy?: SchemaAndPolicy): CursorWithNode<MapTree> | undefined;
/**
 * Transforms an input {@link TypedNode} tree to a {@link MapTree}.
 * @param data - The input tree to be converted.
 * If the data is an unsupported value (e.g. NaN), a fallback value will be used when supported,
 * otherwise an error will be thrown.
 *
 * Fallbacks:
 *
 * * `NaN` =\> `null`
 *
 * * `+/-∞` =\> `null`
 *
 * * `-0` =\> `+0`
 *
 * @param allowedTypes - The set of types allowed by the parent context. Used to validate the input tree.
 * @param context - An optional context which, if present, will allow defaults to be created by {@link ContextualFieldProvider}s.
 * If absent, only defaults from {@link ConstantFieldProvider}s will be created.
 * @param schemaValidationPolicy - The stored schema and policy to be used for validation, if the policy says schema
 * validation should happen. If it does, the input tree will be validated against this schema + policy, and an error will
 * be thrown if the tree does not conform to the schema. If undefined, no validation against the stored schema is done.
 * @remarks The resulting tree will be populated with any defaults from {@link FieldProvider}s in the schema.
 */
export declare function mapTreeFromNodeData(data: InsertableContent, allowedTypes: ImplicitAllowedTypes, context?: NodeKeyManager, schemaValidationPolicy?: SchemaAndPolicy): MapTree;
export declare function mapTreeFromNodeData(data: InsertableContent | undefined, allowedTypes: ImplicitAllowedTypes, context?: NodeKeyManager, schemaValidationPolicy?: SchemaAndPolicy): MapTree | undefined;
/**
 * @returns all types for which the data is schema-compatible.
 */
export declare function getPossibleTypes(allowedTypes: ReadonlySet<TreeNodeSchema>, data: ContextuallyTypedNodeData): TreeNodeSchema[];
/**
 * Content of a tree which needs external schema information to interpret.
 *
 * This format is intended for concise authoring of tree literals when the schema is statically known.
 *
 * Once schema aware APIs are implemented, they can be used to provide schema specific subsets of this type.
 */
export type ContextuallyTypedNodeData = ContextuallyTypedNodeDataObject | number | string | boolean | null | readonly ContextuallyTypedNodeData[];
/**
 * Content of a field which needs external schema information to interpret.
 *
 * This format is intended for concise authoring of tree literals when the schema is statically known.
 *
 * Once schema aware APIs are implemented, they can be used to provide schema specific subsets of this type.
 */
export type ContextuallyTypedFieldData = ContextuallyTypedNodeData | undefined;
/**
 * Object case of {@link ContextuallyTypedNodeData}.
 */
export interface ContextuallyTypedNodeDataObject {
    /**
     * The type of the node.
     * If this node is well-formed, it must follow this schema.
     */
    readonly [typeNameSymbol]?: string;
    /**
     * Fields of this node, indexed by their field keys.
     *
     * Allow explicit undefined for compatibility with FlexTree, and type-safety on read.
     */
    [key: FieldKey]: ContextuallyTypedFieldData;
    /**
     * Fields of this node, indexed by their field keys as strings.
     *
     * Allow unbranded field keys as a convenience for literals.
     */
    [key: string]: ContextuallyTypedFieldData;
}
//# sourceMappingURL=toMapTree.d.ts.map